%% Demo-Programs for the paper ''On multifrequency phase retrieval from noisy data,''
%%  Vladimir Katkovnik and Karen Egiazarian

%% SPAR (WRAPPED and UNWRAPPED) PHASE and AMPLITUDE RECONSTRUCTION  
%% 
%% Version for Gaussian noise
%% Allows subsampling of data random and deterministic
%% Fourier transform propagation
%% Diffraction pattern observation with phase modulation at the obejct plane


%% V. Katkovnik, JANUARY 2018, TUT

close all
clear all
addpath .\DATA
addpath .\PUMA
addpath .\BM3D

%% PARAMETERS of BM3D-FRAME FILTER
Nstep = 3;  N11=8;  N22=8; 
threshType='h'; 
if 0
%% for undersampled data 
threshold_ampl=1.4*4;
threshold_phi=1.4*4
else
%% for full sensor data
threshold_ampl=1.4*2;
threshold_phi=1.4*2;
end
%% PARAMETERS of ALGORITHMS
IterNum=30;                %% ITERATION NUMBER , 
                           %% h_filtering is on 
Fienup=0;                  %%  1 means that the Fienup rules is used instead of the optimal solution for data denoising
filtering=1;               %%  1 for BM3D filtering
if_figure=0;
unwrapping=1;
if_unwrap=0;
if_bias=1;                 %% if if_bias==1, RRMSE is calculated with the shift compensation
h_scale=1.;                %% scale for radians
random_mask=1;             %% identity mask for random_mask=0;
%% TWF=1;                   %% Truncated Wirtinger Flow, TWF=1, acitivation, 0 -deactivation
support_part=1; %%.4       %% Percentage of active pixels is support_part^2
                         %% support_part=.5 gives 25% 
P=1                     %% for random sampling, percentag of active (non-zero) pixels                       

L = 12/12;                  %% Number of Experiments
SNR_set=[5 20 30 40 50];
SNR_set=50;
alpha_2=0.2
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%                      
 %% Complex-valued Object Design
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
test_phase =  4;   %% Phase image selection
switch(test_phase)
   case 0
        varphi=ones(100)*pi/2/2;
 
   case 1
     load DEMO_1                      %% Gaussian abs phase
     varphi=varphi;
   case 3
     load DEMO_3                      %% Truncated Gaussian abs phase
     varphi=varphi;
   case 4
     load DEMO_4_small                 %% Shear plane abs phase
     varphi=varphi;
   case 5
        varphi=double(imread('image_Cameraman256.png'))/255;
        varphi=varphi(60:60+99,60:60+99);
   case 6
        varphi=double(imread('image_Lena256.png'))/255;
        varphi=varphi(60:60+99,60:60+99);
       
   case 7 
        varphi=double(imread('image_Cheese128.png'))/255;
        varphi=imresize(varphi,[256 256],'nearest');
   case 8 
        varphi=double(imread('image_Testpat256.png'))/255*20;
    case 9
        varphi=mean(double(imread('USAF_test.png')),3)/255;
       varphi=varphi(60:60+611,340:340+611)*30;
       %% varphi=varphi(60+100:60+99+100,340+100:340+99+100)*20;
       %% varphi=varphi(60+100:60+49+100,340+100:340+49+100)*20;
    case 10
        load image_Peppers256
        varphi=image_Peppers256;
       
end
varphi=varphi*h_scale;
%%[yN,xN]=size(varphi); 
 
[yN,xN]=size(varphi); 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Amplitude image selection
test_ampl =1;
switch(test_ampl)
   case 1
   %%  Bo=ones(yN,xN); %% Unit amplitude
    Bo=ones(size(varphi));                           %% Unit amplitude
   case 2
     [X,Y] = meshgrid(-xN/2:xN/2-1,-yN/2:yN/2-1);
     Bo=2-((X/(xN/2)).^2+(Y/(yN/2)).^2)+1;           %% Quadratic amplitude
   case 5
     Bo=double(imread('image_Cameraman256.png'))/255+1;
        Bo=Bo(60:60+99,60:60+99)+1;
   case 6
        Bo=double(imread('image_Lena256.png'))/255;
%          Bo=double(imread('image_Lena256.png'));
         Bo=Bo(60:60+99,60:60+99)+1;
   case 7 
        Bo=double(imread('image_Cheese128.png'))/255+1;
       
       %% Bo=imresize(Bo,[256 256],'nearest');
   case 8 
        Bo=double(imread('image_Testpat256.png'))/255+1;
    case 9
        load image_Peppers256
        Bo=image_Peppers256;
end

[n1,n2]=size(varphi);
%% COMPLEX-VALUED OBJECT MODEL 
    x=Bo.*exp(1j*varphi);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% OBSERVATION MODEL and OPERATORS %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Masks   = zeros(yN,xN,L);        %% Storage for L masks, each of dim yN x xN
m       = yN*xN*L;                %% Data length

rand('seed',10001), randn('seed',1001);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% GENERATION of RAMDOM PHASE MODULATION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if 1
    %% Four random values for Phase Mask
for ll = 1:L, Masks(:,:,ll) =randsrc(yN, xN,[1i -1i 1 -1]); end  %% As in TWF
% Sample phases: each symbol in alphabet {1, -1, i , -i} has equal prob.
if random_mask==0
    Masks=ones(size(Masks));
end
else
 %% Gaussian Phase Mask
for ll = 1:L, Masks(:,:,ll) = randn(yN, xN); end
Masks=exp(1j*Masks*pi*random_mask);

end
 %%%%%%%%%%%%%%
 Masks_Support=zeros(size(varphi));

% summetric support
% support_y=[max(-floor(yN/2*support_part)+yN/2,1):min(+floor(yN/2*support_part)+yN/2,yN)];
% support_x=[max(-floor(xN/2*support_part)+xN/2,1):min(+floor(xN/2*support_part)+xN/2,yN)];
[NN,MM]=size(varphi);

%% non-symmetric support from the upper left corner
support_y=[1:floor(NN*support_part)];support_x=support_y;
for ll = 1:L, Masks_Support(support_y,support_x,ll) = 1; end
Masks_Support=Masks_Support.*(rand(size(Masks_Support))<=P);

  %%%%%%%%%%%%%%%%%%%
 %%%%%%%%%%%%%%%%%%%

%% Linear operators for forward and back propagation from  
%% Forward propagation from Y. Chen and E. J. Cands, 
%% "Solving random quadratic systems of equations is nearly as easy as solving linear systems," Communications on Pure and Applied Mathematics 70(5), pp. 822--883, 2017.

%% Forward propagation
A = @(I)  fft2(conj(Masks) .* reshape(repmat(I,[1 L]), size(I,1), size(I,2), L));  % Input is yN x xN image, output is yN x xN x L array

%% Backward propagation
At = @(Y) sum(Masks .* ifft2(Y), 3)/L;                        %% Input is yN x xN x L array, output is yN x xN image
y=zeros(n1,n2,L);
%% TRUE INTENSITIES
y= abs(A(squeeze(x))).^2.* Masks_Support; 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
randn('seed',1001);
s_SNR=0;

for SNR1=SNR_set      %% Loop for SNR values
    tic
s_SNR=s_SNR+1
noise=randn(size(y));
Npower  = mean(y(:).^2)/(10^(SNR1/10));
ssigma=sqrt(Npower);
z=y+ssigma*noise;
z=z.*(z>=0);
gamma_1=ssigma^2/100;      

SNR(s_SNR)=10*log10(sum(y(:).^2)/sum(y(:)-z(:)).^2) %% SNR for Poisson Data
%%%%%%%%%%%%%% SUPPORT SENSOR %%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% INITIALIZATION FOR SPAR
V=varphi;                                         %% for size only
V=ones(size(V)).* exp(1j*pi*randn(size(V))*0.1);  %% RANDOM PHASE INITIALIZATION
V_init=V;
RMSE_ampl(1,s_SNR)=norm(abs(x)-abs(V_init),'fro')/sqrt(yN*xN);  %% RMSE Initial ampl error

%% Phase unwrapping
if unwrapping==1
potential.quantized = 'no';
potential.threshold = pi;
verbose='no';
phi_Uo_init=puma_ho(angle(V_init),.5,'verbose',verbose); %% for demo unwrapping from raw data
bb_init=mean(varphi(:) - phi_Uo_init(:));
RMSE_phi(1,s_SNR) = norm(varphi - phi_Uo_init-bb_init, 'fro')/sqrt(yN*xN); %% RMSE Initial rel. phase error

else
    V_cor_init=exp(-1i*angle(trace(x'*V_init))) * V_init;   %% PHASE CORRECTED INITIAL GUESS 
    bb1_init=-angle(trace(x'*V_init));
    phi_Uo_init=angle(V_init);
    Uo_corr_init=exp(-1i*angle(trace(x'*V_init))) * V_init;   %% CORRECTED INITIAL GUESS 
   %% RMSE_phi(s_KAPPA,1) = norm(angle(x) - angle(Uo_corr_init), 'fro')/sqrt(yN*xN); % EMSE Initial rel. phase error
    RMSE_phi(1,s_SNR) = norm(wrap(angle(x) - angle(Uo_corr_init)), 'fro')/sqrt(yN*xN); % EMSE Initial rel. phase error
end

%% RMSE of the phase and amplitude estimates for use in BM3D filtered estimates
sigma_phi(1)=function_stdEst2D(phi_Uo_init,2);
sigma_ampl(1)=function_stdEst2D(abs(V),2);
%%%%%%%%%%% MAIN ITERATIONS %%%%%%%%%%%%%%%
gamma_10=gamma_1;
for ss=1:IterNum
    gamma_1=gamma_10/(1+0.1*ss);
    if ss<IterNum-8
        alpha=0;
    else
        alpha=alpha_2;
    end
             
ss0=ss;       %% ss0 is iteration when BM3D grouping is updated
%% STEP 1:     %% Forward propagation 

 Us = A(squeeze(V)); 

 %% STEP 2:     %% Gaussian Filtering
if Fienup==1   %% No Filtering at Sensor Plane
Us=(sqrt(squeeze(z)).*Masks_Support+(abs(Us).^1).*(1-Masks_Support)).*exp(j*angle(Us));

else
[Us]=GaussianNoiseFilterCardan(Us,z,ssigma, Masks_Support,gamma_1,L);
end

%% STEP 3: Backward propagation

  Uo=(1-alpha)*At(Us)+alpha*V;  
 
%% STEP 4: Phase unwrapping ;
if unwrapping==1
potential.quantized = 'no';
potential.threshold = pi;
verbose='no';
phi_Uo=puma_ho(angle(Uo),.5,'verbose',verbose); %% for demo unwrapping from raw data

else
    
phi_Uo=double(angle(Uo));
end   
    
%% STEP 5: BM3D phase and amplitude filtering
%% BM3D filtering of phase

sigma_phi(ss+1)=function_stdEst2D(phi_Uo,2);  %% STD of phase
phi_u0_SPAR=BM3D_SPAR_UNWRAP_PHIp(phi_Uo,threshType, sigma_phi(ss+1)*threshold_phi, N11, N22, Nstep,filtering,ss,ss0);


%% BM3D filtering of amplitude           
sigma_ampl(ss+1)=function_stdEst2D(abs(Uo),2);  %% STD of amplitude

B_u0_SPAR=BM3D_SPAR_ABSp(abs(Uo),threshType, sigma_ampl(ss+1)*threshold_ampl, N11, N22, Nstep,filtering,ss,ss0);

 %% STEP 6: UPDATE of x

V=B_u0_SPAR.*exp(1j*phi_u0_SPAR);

V_cor=exp(-1i*(angle(trace(x'*V)))) * V;  %% Phase correction by an invariant shift according to x

if unwrapping==1
bb=mean(varphi(:) - phi_u0_SPAR(:));                                       %% absolute phase shift
RMSE_phi(ss+1,s_SNR) = norm(varphi - phi_u0_SPAR-bb, 'fro')/sqrt(yN*xN); %% RMSE absolute phase error

else
    
RMSE_phi(ss+1,s_SNR) = norm(wrap(angle(x) - (angle(V_cor))), 'fro')/sqrt(yN*xN); %% RMSE phase error
end

RMSE_ampl(ss+1,s_SNR)=norm(abs(x)-abs(V),'fro')/sqrt(yN*xN);                      %% RMSE amplitude error

figure(1), subplot(1,2,1), plot(RMSE_phi,'-o'), grid on, title('RMSE-PHASE'),...
 subplot(1,2,2),   plot( RMSE_ampl,'-s'),grid on, title('RMSE-AMPL')
pause(.1)
if 0   %% ITERATION SHOW
if unwrapping==1
figure(3), subplot(2,2,1), imshow(phi_Uo+bb,[]), title(['IterNum=', num2str(ss+1),'; RMSE-abs.phase=', num2str(RMSE_phi(ss+1,s_SNR),3)]),...
    subplot(2,2,2), imshow(abs(V),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1,s_SNR),3)])
    subplot(2,2,3), imshow(phi_Uo_init+bb_init,[]), title(['IterNum=', num2str(0),'; RMSE-abs.phase=', num2str(RMSE_phi(1,s_SNR),3)]),...
    subplot(2,2,4), imshow(abs(V_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1,s_SNR),3)])
    else
 figure(3), subplot(2,2,1), imshow(angle(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-phase=', num2str(RMSE_phi(ss+1,s_SNR),3)]),...
    subplot(2,2,2), imshow(abs(V_cor),[]), title([' RMSE-ampl=', num2str(RMSE_ampl(ss+1,s_SNR),3)])
    subplot(2,2,3), imshow(angle(V_cor_init),[]), title(['Init-Phase', num2str(0),'; RMSE-phase=', num2str(RMSE_phi(1,s_SNR),3)]),...
    subplot(2,2,4), imshow(abs(V_cor_init),[]), title(['Init-Ampl=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1,s_SNR),3)])
    end
end

end
SPAR_time(s_SNR)=toc    %% for SPAR

if 1
if unwrapping==1
figure(3), subplot(2,2,1), imshow(phi_Uo+bb,[]), title(['IterNum=', num2str(ss+1),'; RMSE-phase=', num2str(RMSE_phi(ss+1,s_SNR),3)]),...
subplot(2,2,2), surfl(varphi), shading interp;colormap(gray);   title(['TRUE ABS PHASE']), VV=axis;,...
subplot(2,2,3), surfl(phi_Uo+bb);  axis=VV; shading interp;colormap(gray);   title([' RECON PHASE, SNR=' num2str(SNR(s_SNR),3),' dB']),...
subplot(2,2,4), imshow(abs(V),[]), title(['RMSE-ampl=', num2str(RMSE_ampl(ss+1,s_SNR),3)])
    else
 figure(31), subplot(2,2,1), imshow(angle(V_cor),[]), title(['IterNum=', num2str(ss+1,s_SNR),'; RMSE-phase=', num2str(RMSE_phi(ss+1,s_SNR),3)]),...
 subplot(2,2,2), imshow(abs(V_cor),[]), title(['RMSE-ampl=', num2str(RMSE_ampl(ss+1,s_SNR),3)])
 subplot(2,2,3), imshow(angle(V_cor_init),[]), title([' RMSE-phase=', num2str(RMSE_phi(1,s_SNR),3)]),...
 subplot(2,2,4), imshow(abs(V_cor_init),[]), title([' RMSE-ampl=', num2str(RMSE_ampl(1,s_SNR),3)])
end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% GS ALGORITHM %%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
end  %% SNR_set



