%% DEMO_SPARSE_PHASE_RETRIEVAL_POISSONIAN.m

%% SPAR (WRAPPED and UNWRAPPED) PHASE and AMPLITUDE RECONSTRUCTION for Phase Retrieval 
%% Random Phase Coded Aperture
%% Demo-Programs for the paper 
%% V. Katkovnik, 'Phase retrieval from noisy data based on sparse approximation of object phase and amplitude'
%% 2017
%% arXiv.org > cs > arXiv:1709.01071


%% V. Katkovnik,  JANUARY 2018,TUT, Finland

close all
clear all
addpath .\DATA
addpath .\PUMA
addpath .\BM3D

tic
%% PARAMETERS of BM3D-FRAME FILTER
Nstep = 3;  N11=8;  N22=8; 
threshType='h'; 
% threshold_ampl=1.4;
% threshold_phi=1.4;
if 1
%% for undersampled data 
threshold_ampl=1.4*4;
threshold_phi=1.4*4
else
%% for full sensor data
threshold_ampl=1.4*2;
threshold_phi=1.4*2;
end



%% PARAMETERS of ALGORITHMS
IterNum=50;                %% ITERATION NUMBER 

unwrapping=1;            %%  1 -means unwrapping on before BM3D phase filtering, 0- means unwrapping off
Fienup=0;                %%  1 means that the Fienup rules is used instead of the optimal solution for data denoising
filtering=1;             %%  1 for BM3D filtering

TWF=1;                   %% Truncated Wirtinger Flow, TWF=1, acitivation, 0 -deactivation
support_part=1;          %% Percentage of active pixels is support_part^2
                         %% support_part=.5 gives 25% 
P=1;                      %% for random sampling, percentag of active (non-zero) pixels                       

L = 12;                  %% Number of Experiments
 KAPPA_set=[.0001/4 .0001/2 .0001 .001 .01 .1 1]; %% set of KAPPA for Poissonian noise scaling
%%KAPPA_set=[ .0001 .001 .01 .1 1]; %% set of KAPPA for Poissonian noise scaling


KAPPA_set=[.001 ];
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%                      
 %% Complex-valued Object Design
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
test_phase = 1;     %% Phase image selection
switch(test_phase)
   case 0
        varphi=zeros(256);
        %% The examples 1,3,4 require the phase unwrapping !!!!
   case 1
     load DEMO_1                      %% Gaussian abs phase
     varphi=varphi;
   case 3
     load DEMO_3                      %% Truncated Gaussian abs phase
     varphi=varphi;
   case 4
     load DEMO_4_small                 %% Shear plane abs phase
     varphi=varphi;
   case 5
        varphi=double(imread('image_Cameraman256.png'))/255*2*pi/4;
   case 6
        varphi=double(imread('image_Lena256.png'))/255*2*pi/4;
   case 7 
        varphi=double(imread('image_Cheese128.png'))/255*2*pi/4;
        varphi=imresize(varphi,[256 256],'nearest');
   case 8 
        varphi=double(imread('image_Testpat256.png'))/255*2*pi/4;
        
end
[yN,xN]=size(varphi); 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Amplitude image selection
test_ampl = 1;
switch(test_ampl)
   case 1
   %%  Bo=ones(yN,xN); %% Unit amplitude
    Bo=ones(size(varphi));                           %% Unit amplitude
   case 2
     [X,Y] = meshgrid(-xN/2:xN/2-1,-yN/2:yN/2-1);
     Bo=2-((X/(xN/2)).^2+(Y/(yN/2)).^2)+.1;           %% Quadratic amplitude
   case 5
        Bo=double(imread('image_Cameraman256.png'))/255+.1;
   case 6
        Bo=double(imread('image_Lena256.png'))/255+.1;
   case 7 
        Bo=double(imread('image_Cheese128.png'))/255+.1;
       
        Bo=imresize(Bo,[256 256],'nearest');
   case 8 
        Bo=double(imread('image_Testpat256.png'))/255+.1;
end
  
x=Bo.*exp(1j*varphi);                           %% TRUE COMPLEX-VALUED OBJECT

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% OBSERVATION MODEL and OPERATORS %%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Masks   = zeros(yN,xN,L);        %% Storage for L masks, each of dim yN x xN
m       = yN*xN*L;                %% Data length

rand('seed',10001), randn('seed',1001);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%% GENERATION of RAMDOM PHASE MODULATION
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
if 1
    %% Four random values for Phase Mask
for ll = 1:L, Masks(:,:,ll) = randsrc(yN, xN,[1i -1i 1 -1]); end  %% As in TWF
% Sample phases: each symbol in alphabet {1, -1, i , -i} has equal prob.
else
    
    %% Gaussian Phase Mask
for ll = 1:L, Masks(:,:,ll) = randn(yN, xN); end
Masks=exp(1j*Masks*pi);
end
 
if 0                          %% small magnitude integer random mask
maxim=5;
for ll = 1:L, Masks1(:,:,ll) = randi((2*maxim)+1,[yN xN])-(1+maxim); end

Masks=exp(1j*Masks1*2*pi/(2*maxim+1));
end


 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%% SUPPORT MASKS and RANDOM SAMPLING %%%%%%%%%%%%%%%%%%
%% CENTRAL PART SUPPORT %%%%%%%%%%%%%
% Masks_Support=zeros(size(varphi));
% support_y=[max(-floor(yN/2*support_part)+yN/2,1):min(+floor(yN/2*support_part)+yN/2,yN)];
% support_x=[max(-floor(xN/2*support_part)+xN/2,1):min(+floor(xN/2*support_part)+xN/2,yN)];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% LEFT UPPER CORNER SUPPORT 
Masks_Support=zeros(size(varphi));
[NN,MM]=size(varphi);
support_y=[1:floor(NN*support_part)];support_x=support_y;


for ll = 1:L, Masks_Support(support_y,support_x,ll) = 1; end

Masks_Support=Masks_Support.*(rand(size(Masks_Support))<=P);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% Linear operators for As 
%% Forward propagation
A = @(I)  fft2(conj(Masks) .* reshape(repmat(I,[1 L]), size(I,1), size(I,2), L));  % Input is yN x xN image, output is yN x xN x L array

%% Backward propagation
At = @(Y) sum(Masks .* ifft2(Y), 3)/L;                        %% Input is yN x xN x L array, output is yN x xN image

At0 = @(Y) sum(Masks .* ifft2(Y), 3) * size(Y,1) * size(Y,2); %% Backward for initialization only

y = abs(A(x)).^2;                                             %% True intensity at the sensor plane
%%%%

s_KAPPA=0;

for KAPPA=KAPPA_set      %% Loop for KAPPA values
s_KAPPA=s_KAPPA+1
gamma_1=1/KAPPA;         %% Parameter of the iterative algorithm


%% Scaled Poissonian data modeling z

rand('seed',100001), randn('seed',1001);
z  = poissrnd(y*KAPPA).*Masks_Support;                %% Poissonain Intensity Observations
y1=y.*Masks_Support;  
Number_Photons_per_Pixel(s_KAPPA)=mean(z(:))                        %% Mean umber of Photons per Pixel
SNR(s_KAPPA)=10*log10(KAPPA^2*sum(y1(:).^2)/sum(y1(:)*KAPPA-z(:)).^2) %% SNR for Poisson Data
%%%%%%%%%%%%%% SUPPORT SENSOR %%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% TWF algorithm %%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if TWF==1
    
    %% TWF algorithm from 
    %% 'Solving Quadratic Systems of Equations via Truncated Wirtinger Flow,' Yuxin Chen and Emmanuel J. Cands, 2015
    %% codes available at http://web.stanford.edu/~yxchen/TWF/

%% INITIALIZATION of ALGORITHM
if exist('Params')                == 0,  Params.n1          = 100;  end
if isfield(Params, 'n2')          == 0,  Params.n2          = 100;  end             % signal dimension
if isfield(Params, 'L')           == 0,  Params.L           = 12;   end             % number of measurements
if isfield(Params, 'grad_type')   == 0,  Params.grad_type   = 'TWF_Poiss';  end     % 'TWF_Poiss': Poisson likelihood

if isfield(Params, 'alpha_lb')    == 0,  Params.alpha_lb    = 0.3;  end
if isfield(Params, 'alpha_ub')    == 0,  Params.alpha_ub    = 5;    end
if isfield(Params, 'alpha_h')     == 0,  Params.alpha_h     = 5;    end
if isfield(Params, 'alpha_y')     == 0,  Params.alpha_y     = 3;    end 
if isfield(Params, 'T')           == 0,  Params.T           = 100/2; end    	% number of iterations
if isfield(Params, 'mu')          == 0,  Params.mu          = 0.2;  end		% step size / learning parameter
if isfield(Params, 'npower_iter') == 0,  Params.npower_iter = 50;   end		% number of power iterations

Params.n1=yN;
Params.n2=xN; Params.L=L; Params.m=m;

tic       %% for TWF

[Relerrs,RMSE_phase_TWF, RMSE_ampl_TWF,x_hat,x_hat_0,bb1_TWF] = TWF_my(z.*Masks_Support,x, Params, A, At0,KAPPA); %% Truncated WF algorithm
RMSE_phi_TWF_set(s_KAPPA)=RMSE_phase_TWF(end);
RMSE_ampl_TWF_set(s_KAPPA)=RMSE_ampl_TWF(end);


time_TWF(s_KAPPA)=toc  %% for TWF


if 0
fprintf('Relative error after initialization: %f\n', Relerrs(1))
fprintf('Relative error after %d iterations: %f\n', T, Relerrs(T+1))
 
figure(1), semilogy(0:Params.T,Relerrs) 
xlabel('Iteration'), ylabel('Relative error (log10)'), ...
     title('Relative error vs. iteration count'), grid on
 
 figure(2), plot(0:Params.T,RMSE_phase_TWF) 
xlabel('Iteration'), ylabel('Relative RMSE'), ...
     title('Phase RMSE vs. iteration count'), grid on
end
%% Visualization for TWF

if 0
figure(1), plot(0:Params.T,Relerrs),... %% semilogy(0:Params.T,Relerrs) 
xlabel('Iteration'), ylabel('Relative Error'),grid on, ...
title('Relative error vs. iteration count')
figure(2), subplot(2,2,1), imshow(angle(x),[]),title('phase-x'),...
subplot(2,2,2), imshow(angle(x_hat),[]),title('phase-x-TWF'),
subplot(2,2,3), imshow(abs((x)),[]),title('ampl-x'),...
subplot(2,2,4), imshow(abs(x_hat),[]),title('ampl-x-TWF'),

end

end   %% end of TWF


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%% SPAR algorithm %%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tic                 %% for SPAR
%% INITIALIZATION FOR SPAR
V=varphi;                                         %% for size only
V=ones(size(V)).* exp(1j*pi*randn(size(V))*0.1);  %% RANDOM PHASE INITIALIZATION
V_init=V;
RMSE_ampl(s_KAPPA,1)=norm(abs(x)-abs(V_init),'fro')/sqrt(yN*xN);  %% RMSE Initial ampl error

%% Phase unwrapping
if unwrapping==1
potential.quantized = 'no';
potential.threshold = pi;
verbose='no';
phi_Uo_init=puma_ho(angle(V_init),.5,'verbose',verbose); %% for demo unwrapping from raw data
bb_init=mean(varphi(:) - phi_Uo_init(:));
RMSE_phi(s_KAPPA,1) = norm(varphi - phi_Uo_init-bb_init, 'fro')/sqrt(yN*xN); %% RMSE Initial rel. phase error

else
    V_cor_init=exp(-1i*angle(trace(x'*V_init))) * V_init;   %% PHASE CORRECTED INITIAL GUESS 
    bb1_init=-angle(trace(x'*V_init));
    phi_Uo_init=angle(V_init);
    Uo_corr_init=exp(-1i*angle(trace(x'*V_init))) * V_init;   %% CORRECTED INITIAL GUESS 
   %% RMSE_phi(s_KAPPA,1) = norm(angle(x) - angle(Uo_corr_init), 'fro')/sqrt(yN*xN); % EMSE Initial rel. phase error
    RMSE_phi(s_KAPPA,1) = norm(wrap(angle(x) - angle(Uo_corr_init)), 'fro')/sqrt(yN*xN); % EMSE Initial rel. phase error
end

%% RMSE of the phase and amplitude estimates for use in BM3D filtered estimates
sigma_phi(1)=function_stdEst2D(phi_Uo_init,2);
sigma_ampl(1)=function_stdEst2D(abs(V),2);

%% MAIN ITERATIONS of SPAR PHASE RETRIEVAL
gamma_10=gamma_1;
for ss=1:IterNum
             
ss0=ss;       %% ss0 is iteration when BM3D grouping is updated
%% STEP 1:     %% Forward propagation 
Us=A(V);       

%% STEP 2:     %% Poissonian Filtering
if Fienup==1   %% No Filtering at Sensor Plane

    Us=(sqrt(((z).*Masks_Support/KAPPA+((abs(Us)).^2).*(1-Masks_Support)))).*exp(j*angle(Us));
else
    [Us]=NoiseFiltering_POISSONIAN_JAN_2018(Us,z,Masks_Support,KAPPA,gamma_1,L);  %% Filtering at Sensor Plane
end


%% STEP 3: Backward propagation

Uo=At(Us);  

%% STEP 4: Phase unwrapping ;
if unwrapping==1
potential.quantized = 'no';
potential.threshold = pi;
verbose='no';
phi_Uo=puma_ho(angle(Uo),.5,'verbose',verbose); %% for demo unwrapping from raw data

else
    
phi_Uo=double(angle(Uo));
end   
    
%% STEP 5: BM3D phase and amplitude filtering
%% BM3D filtering of phase

sigma_phi(ss+1)=function_stdEst2D(phi_Uo,2);  %% STD of phase
phi_u0_SPAR=BM3D_SPAR_UNWRAP_PHIp(phi_Uo,threshType, sigma_phi(ss+1)*threshold_phi, N11, N22, Nstep,filtering,ss,ss0);


%% BM3D filtering of amplitude           
sigma_ampl(ss+1)=function_stdEst2D(abs(Uo),2);  %% STD of amplitude

B_u0_SPAR=BM3D_SPAR_ABSp(abs(Uo),threshType, sigma_ampl(ss+1)*threshold_ampl, N11, N22, Nstep,filtering,ss,ss0);

 %% STEP 6: UPDATE of x

V=B_u0_SPAR.*exp(1j*phi_u0_SPAR);

V_cor=exp(-1i*(angle(trace(x'*V)))) * V;  %% Phase correction by an invariant shift according to x

if unwrapping==1
bb=mean(varphi(:) - phi_u0_SPAR(:));                                       %% absolute phase shift
RMSE_phi(s_KAPPA,ss+1) = norm(varphi - phi_u0_SPAR-bb, 'fro')/sqrt(yN*xN); %% RMSE absolute phase error

else
    
RMSE_phi(s_KAPPA,ss+1) = norm(wrap(angle(x) - (angle(V_cor))), 'fro')/sqrt(yN*xN); %% RMSE phase error
end

RMSE_ampl(s_KAPPA,ss+1)=norm(abs(x)-abs(V),'fro')/sqrt(yN*xN);                      %% RMSE amplitude error

% figure(3), subplot(2,2,1), imshow(angle(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-phase=', num2str(RMSE_phi(ss+1),3)]),...
%     subplot(2,2,2), imshow(abs(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1),3)])
%     subplot(2,2,3), imshow(angle(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-phase=', num2str(RMSE_phi(1),3)]),...
%     subplot(2,2,4), imshow(abs(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1),3)])
% 
if 0   %% ITERATION SHOW
if unwrapping==1
figure(3), subplot(2,2,1), imshow(phi_Uo+bb,[]), title(['IterNum=', num2str(ss+1),'; RMSE-abs.phase=', num2str(RMSE_phi(ss+1),3)]),...
    subplot(2,2,2), imshow(abs(V),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1),3)])
    subplot(2,2,3), imshow(phi_Uo_init+bb_init,[]), title(['IterNum=', num2str(0),'; RMSE-abs.phase=', num2str(RMSE_phi(1),3)]),...
    subplot(2,2,4), imshow(abs(V_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1),3)])
    else
 figure(3), subplot(2,2,1), imshow(angle(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-abs.phase=', num2str(RMSE_phi(ss+1),3)]),...
    subplot(2,2,2), imshow(abs(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1),3)])
    subplot(2,2,3), imshow(angle(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-abs.phase=', num2str(RMSE_phi(1),3)]),...
    subplot(2,2,4), imshow(abs(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1),3)])
    end
end
%%[ss RMSE_phi(s_KAPPA,ss) RMSE_ampl(s_KAPPA,ss)]


end
SPAR_time(s_KAPPA)=toc    %% for SPAR

if 0
if unwrapping==1
figure(3), subplot(2,2,1), imshow(phi_Uo+bb,[]), title(['IterNum=', num2str(ss+1),'; RMSE-abs.phase=', num2str(RMSE_phi(ss+1),3)]),...
    subplot(2,2,2), imshow(abs(V),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1),3)])
    subplot(2,2,3), imshow(phi_Uo_init+bb_init,[]), title(['IterNum=', num2str(0),'; RMSE-abs.phase=', num2str(RMSE_phi(1),3)]),...
    subplot(2,2,4), imshow(abs(V_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1),3)])
    else
 figure(3), subplot(2,2,1), imshow(angle(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-abs.phase=', num2str(RMSE_phi(ss+1),3)]),...
    subplot(2,2,2), imshow(abs(V_cor),[]), title(['IterNum=', num2str(ss+1),'; RMSE-ampl=', num2str(RMSE_ampl(ss+1),3)])
    subplot(2,2,3), imshow(angle(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-abs.phase=', num2str(RMSE_phi(1),3)]),...
    subplot(2,2,4), imshow(abs(V_cor_init),[]), title(['IterNum=', num2str(0),'; RMSE-ampl=', num2str(RMSE_ampl(1),3)])
    end
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%% GS ALGORITHM %%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
tic                         %% for GS
PHASE_RETRIEVAL_CODED_FOURIER_GS_ALGORITHM
time_GS(s_KAPPA)=toc        %% for GS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%% Images for results %%%%%%
if TWF==1
figure, subplot (1,4,2), imshow(angle(V_cor_GS),[]), title(['GS, RMSE =', num2str(RMSE_phi_GS(s_KAPPA,end),2)]),...
    subplot(1,4,1), imshow(angle(x_hat),[]), title(['TWF,  \chi =  ', num2str(KAPPA,2),', RMSE = ', num2str(RMSE_phi_TWF_set(end),2)]),...
    subplot(1,4,3), imshow(wrap(phi_u0_SPAR),[]), title(['SPAR, RMSE = ', num2str(RMSE_phi(s_KAPPA,end),2)]) 
    subplot(1,4,4), imshow(wrap(varphi),[]), title('TRUE') 
% figure, subplot (1,3,2), imshow(angle(V_cor_GS),[]), title(['GS, RMSE =', num2str(RMSE_phi_GS(s_KAPPA,end),2)]),...
%     subplot(1,3,1), imshow(angle(x_hat),[]), title(['TWF, RMSE = ', num2str(RMSE_phi_TWF_set(end),2)]),...
%     subplot(1,3,3), imshow(wrap(phi_u0_SPAR),[]), title(['SPAR, RMSE = ', num2str(RMSE_phi(s_KAPPA,end),2)]) 
 %%   subplot(1,4,4), imshow(wrap(varphi),[]), title('TRUE') 
pause(.2)
if 0
figure, subplot (1,3,2), imshow(abs(V_GS),[]), title(['GS AMPL, RMSE = ', num2str(RMSE_ampl_GS(s_KAPPA,end),2)]),...
    subplot(1,3,1), imshow(abs(x_hat),[]), title(['TWF AMPL,  \chi =  ', num2str(KAPPA,2),', RMSE = ', num2str(RMSE_ampl_TWF_set(end),2)]),...
    subplot(1,3,3), imshow(abs(V),[]), title(['SPAR AMPL, RMSE = ', num2str(RMSE_ampl(s_KAPPA,end),2)]) 
pause(.2)
end
else
    figure, subplot (1,2,1), imshow(angle(V_cor_GS),[]), title(['GS PHASE, \chi =  ', num2str(KAPPA,2),', RMSE =', num2str(RMSE_phi_GS(s_KAPPA,end),2)]),...
   %% subplot(1,3,1), imshow(angle(x_hat),[]), title(['TWF PHASE, \chi =  ', num2str(KAPPA,2),', RMSE = ', num2str(RMSE_phi_TWF_set(end),2)]),...
    subplot(1,2,2), imshow(angle(V_cor),[]), title(['SPAR PHASE, RMSE = ', num2str(RMSE_phi(s_KAPPA,end),2)]) 
pause(.2)
if 0
figure, subplot (1,2,1), imshow(abs(V_GS),[]), title(['GS AMPL,  \chi =  ', num2str(KAPPA,2),', RMSE = ', num2str(RMSE_ampl_GS(s_KAPPA,end),2)]),...
 %%   subplot(1,3,1), imshow(abs(x_hat),[]), title(['TWF AMPL, \chi =  ', num2str(KAPPA,2),', RMSE = ', num2str(RMSE_ampl_TWF_set(end),2)]),...
        subplot(1,2,2), imshow(abs(V),[]), title(['SPAR AMPL, RMSE = ', num2str(RMSE_ampl(s_KAPPA,end),2)]) 
pause(.2)
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%
end
PLOTS_Combined
end  %% SNR_set
