%% Truncated Wirtinger Flow (TWF) algorithm proposed in the paper
%  ``Solving Random Quadratic Systems of Equations Is Nearly as Easy as Solving Linear Systems'' by Y. Chen and E. J. Cands.
%  The code below is adapted from implementation of the Wirtinger Flow algorithm written by M. Soltanolkotabi and E. Candes

function [Relerrs,RMSE_phase, RMSE_ampl, x_hat,z0,bb1_TWF] = TWF_my(y,x,Params, A, At,KAPPA)    
%% Initialization
%% y-observations
%% x - true
%% z - estimates
    npower_iter = Params.npower_iter;           % Number of power iterations 
    z0 = randn(Params.n1,Params.n2); z0 = z0/norm(z0,'fro');    % Initial guess 
    normest = sqrt(sum(y(:))/numel(y(:)));    % Estimate norm to scale eigenvector  
    
    for tt = 1:npower_iter,                     % Truncated power iterations
        ytr = y.* (abs(y) <= Params.alpha_y^2 * normest^2 );
        z0 = At( ytr.* (A(z0)) ); z0 = z0/norm(z0,'fro');
    end
    
    z = normest * z0;                   % Apply scaling 
    z0=z/sqrt(KAPPA);
    
Relerrs = norm(x - exp(-1i*angle(trace(x'*z))) * z/sqrt(KAPPA), 'fro')/norm(x,'fro'); % Initial rel. error
bb1_TWF=-angle(trace(x'*z));
RMSE_phase = norm(wrap(angle(x) - angle(exp(-1i*angle(trace(x'*z))) * z)),'fro')/sqrt(length(x(:))); % Initial rel. error
%% RMSE_phase = norm(angle(x) - (angle(z)+bb1_TWF),'fro')/sqrt(length(x(:))); % Initial rel. error
RMSE_ampl = norm(abs(x) - abs(z)/sqrt(KAPPA),'fro')/sqrt(length(x(:))); % Initial rel. error


%% Loop
    grad_type = Params.grad_type;
    if strcmp(grad_type, 'TWF_Poiss') == 1
        mu = @(t) Params.mu; % Schedule for step size 
    elseif strcmp(grad_type, 'WF_Poiss') == 1
        tau0 = 330;                         % Time constant for step size
        mu = @(t) min(1-exp(-t/tau0), 0.2); % Schedule for step size  
    end

    for t = 1: Params.T,
        
        grad = compute_grad(z, y, Params, A, At);
       
        z = z - mu(t) * grad;             % Gradient update 
   %%     Relerrs_phase = [Relerrs_phase, norm(angle(x) - angle(exp(-1i*angle(trace(x'*z))) * z/sqrt(KAPPA)), 'fro')/norm(angle(x),'fro')];  
   
 bb1_TWF=-angle(trace(x'*z));
 RMSE_phase = [RMSE_phase,norm(wrap(angle(x) - angle(exp(-1i*angle(trace(x'*z))) * z)),'fro')/sqrt(length(x(:)))]; % Initial rel. error
%% RMSE_phase = [RMSE_phase,norm(angle(x) - (angle(z)+bb1_TWF),'fro')/sqrt(length(x(:)))]; % Initial rel. error
 RMSE_ampl = [RMSE_ampl,norm(abs(x) - abs(z)/sqrt(KAPPA),'fro')/sqrt(length(x(:)))]; 
 Relerrs = [Relerrs, norm(x - exp(-1i*angle(trace(x'*z))) * z/sqrt(KAPPA), 'fro')/norm(x,'fro')];  
   
    end
%%    keyboard
   x_hat=exp(-1i*angle(trace(x'*z))) * z/sqrt(KAPPA);
   %%  x_hat=z/KAPPA;
 
